@@ -2,23 +2,30 @@ from __future__ import annotations
 
 import os
 import platform
+import random
 import re
 import subprocess as sp
 import sys
 import traceback
+import uuid
 from argparse import ArgumentParser
+from contextlib import contextmanager
 from io import StringIO
 from pathlib import Path
-from pathlib import PurePosixPath
 from tempfile import TemporaryDirectory
 from typing import Any
+from typing import Callable
+from unittest.mock import patch
 
 from rich.console import Console
 
+from inline_snapshot._config import Config
+from inline_snapshot._config import read_config
 from inline_snapshot._exceptions import UsageError
-from inline_snapshot._external import DiscStorage
+from inline_snapshot._external._storage._hash import HashStorage
 from inline_snapshot._problems import report_problems
 
+from .._change import ChangeBase
 from .._change import apply_all
 from .._flags import Flags
 from .._global_state import snapshot_env
@@ -48,43 +55,188 @@ def normalize(text):
     # fix windows problems
     text = text.replace("\u2500", "-")
     text = text.replace("\r", "")
-    text = text.replace(" \n", " ⏎\n")
     return text
 
 
+@contextmanager
+def deterministic_uuid():
+    rd = random.Random(0)
+
+    def f():
+        return uuid.UUID(int=rd.getrandbits(128), version=4)
+
+    with patch("uuid.uuid4", new=f):
+        yield
+
+
+# this code is copied from pytest
+
+# Regex to match the session duration string in the summary: "74.34s".
+rex_session_duration = re.compile(r"\d+\.\d\ds")
+# Regex to match all the counts and phrases in the summary line: "34 passed, 111 skipped".
+rex_outcome = re.compile(r"(\d+) (\w+)")
+
+
+def parse_outcomes(lines):
+    for line in reversed(lines):
+        if rex_session_duration.search(line):
+            outcomes = rex_outcome.findall(line)
+            ret = {noun: int(count) for (count, noun) in outcomes}
+            break
+        else:
+            pass  # pragma: no cover
+    else:
+        raise ValueError("Pytest terminal summary report not found")  # pragma: no cover
+
+    to_plural = {
+        "warning": "warnings",
+        "error": "errors",
+    }
+    return {to_plural.get(k, k): v for k, v in ret.items()}
+
+
+conftest_footer = """
+import uuid
+import random
+
+rd = random.Random(0)
+
+def f():
+    return uuid.UUID(int=rd.getrandbits(128), version=4)
+
+uuid.uuid4 = f
+"""
+
+
+@contextmanager
+def change_file(path, map_function):
+    exists = path.exists()
+    if exists:
+        text = path.read_text()
+    else:
+        text = ""
+
+    path.write_text(map_function(text))
+
+    yield
+
+    if exists:
+        path.write_text(text)
+    else:
+        path.unlink()
+
+
 class Example:
-    def __init__(self, files: str | dict[str, str]):
+    files: dict[str, str | bytes]
+
+    def __init__(self, files: str | dict[str, str | bytes]):
         """
         Parameters:
-            files: a collecton of files where inline-snapshot opperates on,
-                   or just a string which will be saved as *test_something.py*.
+            files: a collection of files which are used as your example project,
+                   or just a string which will be saved as *tests/test_something.py*.
         """
         if isinstance(files, str):
-            files = {"test_something.py": files}
+            files = {"tests/test_something.py": files}
 
         self.files = files
 
-        self.dump_files()
-
     def dump_files(self):
+        from rich.panel import Panel
+        from rich.text import Text
+
+        console = Console()
+
         for name, content in self.files.items():
-            print(f"file: {name}")
-            print(content)
-            print()
+            if isinstance(content, bytes):
+                content = repr(content)
+            console.print(Panel(Text(content), title=name))
 
     def _write_files(self, dir: Path):
         for name, content in self.files.items():
             filename = dir / name
             filename.parent.mkdir(exist_ok=True, parents=True)
-            filename.write_text(content)
+            if isinstance(content, str):
+                filename.write_text(content)
+            else:
+                filename.write_bytes(content)
 
     def _read_files(self, dir: Path):
+
+        def try_read(path: Path):
+            try:
+                return path.read_text("utf-8")
+            except UnicodeDecodeError:
+                return path.read_bytes()
+
+        def normalize_path(path):
+            return str(path.relative_to(dir)).replace("\\", "/")
+
         return {
-            str(p.relative_to(dir)): p.read_text()
-            for p in [*dir.iterdir(), *dir.rglob("*.py")]
+            normalize_path(p): try_read(p)
+            for p in dir.rglob("*")
             if p.is_file()
+            and p.name != ".gitignore"
+            and p.suffix != ".pyc"
+            and ".pytest_cache" not in p.parts
         }
 
+    def with_files(self, extra_files: dict[str, str | bytes]) -> Example:
+        """
+        Adds extra files to the example.
+
+        Arguments:
+            extra_files: dictionary of filenames and file contents.
+        """
+        return Example({**self.files, **extra_files})
+
+    def read_file(self, filename: str) -> str:
+        """
+        Reads a file from the example.
+
+        Arguments:
+            filename: the filename.
+        """
+        text = self.files[filename]
+        assert isinstance(text, str)
+        return text
+
+    read_text = read_file
+
+    def change_code(self, mapping: Callable[[str], str]) -> Example:
+        """
+        Changes example tests by mapping every file with the given function.
+
+        Arguments:
+            mapping: function to apply to each file's content.
+        """
+        return Example(
+            {
+                name: mapping(text) if isinstance(text, str) else text
+                for name, text in self.files.items()
+            }
+        )
+
+    def replace(self, old_text: str, new_text: str) -> Example:
+        """
+        Changes example tests by replacing old_text with new_text.
+
+        Arguments:
+            old_text: the text to be replaced.
+            new_text: the new text to use instead.
+        """
+        return self.change_code(lambda code: code.replace(old_text, new_text))
+
+    def remove_file(self, filename: str) -> Example:
+        """
+        Removes a file from the example.
+
+        Arguments:
+            filename: the file to be removed.
+        """
+        return Example(
+            {name: file for name, file in self.files.items() if name != filename}
+        )
+
     def run_inline(
         self,
         args: list[str] = [],
@@ -100,14 +252,14 @@ class Example:
         This is useful for fast test execution.
 
         Parameters:
-            args: inline-snapshot arguments (supports only "--inline-snapshot=fix|create|..." ).
+            args: inline-snapshot arguments (supports only "--inline-snapshot=fix|create|...").
             reported_categories: snapshot of categories which inline-snapshot thinks could be applied.
             changed_files: snapshot of files which are changed by this run.
-            raises: snapshot of the exception which is raised during the test execution.
-                    It is required if your code raises an exception.
+            raises: snapshot of the exception raised during test execution.
+                    Required if your code raises an exception.
 
         Returns:
-            A new Example instance which contains the changed files.
+            A new Example instance containing the changed files.
         """
 
         parser = ArgumentParser()
@@ -128,6 +280,7 @@ class Example:
         )
         parsed_args = parser.parse_args(args)
         flags = (parsed_args.inline_snapshot or "").split(",")
+        self.dump_files()
 
         with TemporaryDirectory() as dir:
             tmp_path = Path(dir)
@@ -135,15 +288,27 @@ class Example:
             self._write_files(tmp_path)
 
             raised_exception = []
-            with snapshot_env() as state:
+
+            with snapshot_env() as state, deterministic_uuid():
+
                 recorder = ChangeRecorder()
                 state.update_flags = Flags({*flags})
-                state.storage = DiscStorage(tmp_path / ".storage")
+                state.active = "disable" not in flags
+                state.all_storages["hash"] = HashStorage(
+                    tmp_path / ".inline-snapshot" / "external"
+                )
+                state.config = Config()
+                read_config(tmp_path / "pyproject.toml", state.config)
+                if state.config.storage_dir is None:
+                    state.config.storage_dir = tmp_path / ".inline_snapshot"
+                else:
+                    pass  # pragma: no cover
+
                 try:
                     tests_found = False
-                    for filename in tmp_path.glob("*.py"):
+                    for filename in tmp_path.rglob("test_*.py"):
                         globals: dict[str, Any] = {}
-                        print("run> pytest", filename)
+                        print("run> pytest-inline", filename)
                         exec(
                             compile(filename.read_text("utf-8"), filename, "exec"),
                             globals,
@@ -169,9 +334,9 @@ class Example:
                 finally:
                     state.active = False
 
-                changes = []
+                changes: list[ChangeBase] = []
                 for snapshot in state.snapshots.values():
-                    changes += snapshot._changes()
+                    changes += list(snapshot._changes())
 
                 snapshot_flags = {change.flag for change in changes}
 
@@ -185,11 +350,15 @@ class Example:
                 )
                 recorder.fix_all()
 
+                for change in changes:
+                    if change.flag in state.update_flags.to_set():
+                        change.apply_external_changes()
+
                 report_output = StringIO()
                 console = Console(file=report_output, width=80)
 
                 # TODO: add all the report output here
-                report_problems(console)
+                report_problems(lambda: console)
 
             if reported_categories is not None:
                 assert sorted(snapshot_flags) == reported_categories
@@ -205,18 +374,27 @@ class Example:
                 assert raises == None
 
             if changed_files is not None:
-                current_files = {}
-
-                for name, content in sorted(self._read_files(tmp_path).items()):
-                    if name not in self.files or self.files[name] != content:
-                        current_files[name] = content
-                assert changed_files == current_files
+                assert changed_files == self._changed_files(tmp_path)
 
             if report is not None:
                 assert report == normalize(report_output.getvalue())
 
             return Example(self._read_files(tmp_path))
 
+    def _changed_files(self, tmp_path):
+        current_files = {}
+        all_current_files = self._read_files(tmp_path)
+
+        for name, content in sorted(all_current_files.items()):
+            if name not in self.files or self.files[name] != content:
+                current_files[name] = content
+
+        for name in self.files:
+            if name not in all_current_files:
+                current_files[name] = None
+
+        return current_files
+
     def run_pytest(
         self,
         args: list[str] = [],
@@ -225,28 +403,33 @@ class Example:
         env: dict[str, str] = {},
         changed_files: Snapshot[dict[str, str]] | None = None,
         report: Snapshot[str] | None = None,
+        error: Snapshot[str] | None = None,
         stderr: Snapshot[str] | None = None,
-        returncode: Snapshot[int] | None = None,
+        returncode: Snapshot[int] = 0,
+        stdin: bytes = b"",
+        outcomes: Snapshot[dict[str, int]] | None = None,
     ) -> Example:
-        """Run pytest with the given args and env variables in an seperate
+        """Run pytest with the given args and environment variables in a separate
         process.
 
-        It can be used to test the interaction between your code and pytest, but it is a bit slower than `run_inline`
+        It can be used to test the interaction between your code and pytest, but it is a bit slower than `run_inline`.
 
         Parameters:
             args: pytest arguments like "--inline-snapshot=fix"
             env: dict of environment variables
-            changed_files: snapshot of files which are changed by this run.
+            changed_files: snapshot of files changed by this run.
             report: snapshot of the report at the end of the pytest run.
             stderr: pytest stderr output
-            returncode: snapshot of the pytest returncode.
+            returncode: snapshot of the pytest return code.
 
         Returns:
-            A new Example instance which contains the changed files.
+            A new Example instance containing the changed files.
         """
+        self.dump_files()
 
         with TemporaryDirectory() as dir:
             tmp_path = Path(dir)
+
             self._write_files(tmp_path)
 
             cmd = [sys.executable, "-m", "pytest", *args]
@@ -257,23 +440,36 @@ class Example:
                 term_columns + 1 if platform.system() == "Windows" else term_columns
             )
             command_env.pop("CI", None)
+            command_env.pop("GITHUB_ACTIONS", None)
+
+            if stdin:
+                # makes Console.is_terminal == True
+                command_env["FORCE_COLOR"] = "true"
 
             command_env.update(env)
 
-            result = sp.run(cmd, cwd=tmp_path, capture_output=True, env=command_env)
+            with change_file(
+                tmp_path / "conftest.py", lambda text: text + conftest_footer
+            ):
+                result = sp.run(
+                    cmd, cwd=tmp_path, capture_output=True, env=command_env, input=stdin
+                )
+
+            result_stdout = result.stdout.decode()
+            result_stderr = result.stderr.decode()
+            result_returncode = result.returncode
 
             print("run>", *cmd)
             print("stdout:")
-            print(result.stdout.decode())
+            print(result_stdout)
             print("stderr:")
-            print(result.stderr.decode())
+            print(result_stderr)
 
-            if returncode is not None:
-                assert result.returncode == returncode
+            assert result.returncode == returncode
 
             if stderr is not None:
 
-                original = result.stderr.decode().splitlines()
+                original = result_stderr.splitlines()
                 lines = [
                     line
                     for line in original
@@ -292,8 +488,8 @@ class Example:
 
                 report_list = []
                 record = False
-                for line in result.stdout.decode().splitlines():
-                    line = line.strip()
+                for line in result_stdout.splitlines():
+                    line = normalize(line.strip())
                     if line.startswith("===="):
                         record = False
 
@@ -308,14 +504,25 @@ class Example:
 
                 report_str = "\n".join(report_list)
 
-                assert normalize(report_str) == report, repr(report_str)
+                assert report_str == report, repr(report_str)
+
+            if error is not None:
+                assert (
+                    error
+                    == "\n".join(
+                        [
+                            line
+                            for line in result_stdout.splitlines()
+                            if line and line[:2] in ("> ", "E ")
+                        ]
+                    )
+                    + "\n"
+                )
 
             if changed_files is not None:
-                current_files = {}
+                assert changed_files == self._changed_files(tmp_path)
 
-                for name, content in sorted(self._read_files(tmp_path).items()):
-                    if name not in self.files or self.files[name] != content:
-                        current_files[str(PurePosixPath(*Path(name).parts))] = content
-                assert changed_files == current_files
+            if outcomes is not None:
+                assert outcomes == parse_outcomes(result_stdout.splitlines())
 
             return Example(self._read_files(tmp_path))
